@@ -554,7 +554,7 @@ static int iwbmp_read_palette(struct iwbmprcontext *rctx)
 static void bmpr_convert_row_32_16(struct iwbmprcontext *rctx, const iw_byte *src, size_t row)
 {
 	int i,k;
-	unsigned int v,x;
+	unsigned int x;
 	int numchannels;
 
 	numchannels = rctx->has_alpha_channel ? 4 : 3;
@@ -567,8 +567,10 @@ static void bmpr_convert_row_32_16(struct iwbmprcontext *rctx, const iw_byte *sr
 		else { // 16
 			x = ((unsigned int)src[i*2+0]) | ((unsigned int)src[i*2+1])<<8;
 		}
-		v = 0;
+
 		for(k=0;k<numchannels;k++) { // For red, green, blue [, alpha]:
+			unsigned int v;
+
 			v = x & rctx->bf_mask[k];
 			if(rctx->bf_low_bit[k]>0)
 				v >>= rctx->bf_low_bit[k];
@@ -708,7 +710,7 @@ static int bmpr_read_rle_internal(struct iwbmprcontext *rctx)
 {
 	int retval = 0;
 	int pos_x, pos_y;
-	iw_byte buf[255];
+	iw_byte buf[256];
 	size_t n_pix;
 	size_t n_bytes;
 	size_t i;
@@ -759,6 +761,7 @@ static int bmpr_read_rle_internal(struct iwbmprcontext *rctx)
 				else {
 					n_bytes = ((n_pix+1)/2)*2;
 				}
+				if(n_bytes>sizeof(buf)) goto done;
 				if(!iwbmp_read(rctx,buf,n_bytes)) goto done;
 				for(i=0;i<n_pix;i++) {
 					if(pos_x<rctx->img->width) {
@@ -847,11 +850,13 @@ static int bmpr_read_rle(struct iwbmprcontext *rctx)
 		!(rctx->compression==IWBMP_BI_RLE4 && rctx->bitcount==4))
 	{
 		iw_set_error(rctx->ctx,"Compression type incompatible with image type");
+		goto done;
 	}
 
 	if(rctx->topdown) {
 		// The documentation says that top-down images may not be compressed.
 		iw_set_error(rctx->ctx,"Compression not allowed with top-down images");
+		goto done;
 	}
 
 	// RLE-compressed BMP images don't have to assign a color to every pixel,
@@ -1279,6 +1284,8 @@ static int iwbmp_calc_bitfields_masks(struct iwbmpwcontext *wctx, int num_masks)
 	int bits[4]; // R, G, B, A
 	int tot_bits = 0;
 
+	iw_zeromem(bits, sizeof(bits));
+	if(num_masks>4) return 0;
 	for(k=0;k<num_masks;k++) {
 		bits[k] = iw_max_color_to_bitdepth(wctx->maxcolor[k]);
 		tot_bits += bits[k];
@@ -1940,7 +1947,7 @@ static int rle_patch_file_size(struct iwbmpwcontext *wctx,size_t rlesize)
 	}
 
 	// Patch the "bits" size
-	ret=(*wctx->iodescr->seek_fn)(wctx->ctx,wctx->iodescr,fileheader_size+20,SEEK_SET);
+	ret=(*wctx->iodescr->seek_fn)(wctx->ctx,wctx->iodescr,(iw_int64)fileheader_size+20,SEEK_SET);
 	if(!ret) return 0;
 	iw_set_ui32le(buf,(unsigned int)rlesize);
 	iwbmp_write(wctx,buf,4);
@@ -2045,7 +2052,9 @@ static int setup_16_32bit(struct iwbmpwcontext *wctx,
 	}
 
 	if(has_alpha && wctx->bmpversion<5) {
-		iw_set_error(wctx->ctx,"Internal: Attempt to write v3 16- or 32-bit image with transparency");
+		// Possible to get here if -bmptrans option was used
+		iw_set_error(wctx->ctx,"Incompatible combination of BMP version, "
+			"transparency, number of colors");
 		return 0;
 	}
 
@@ -2084,7 +2093,17 @@ static int iwbmp_write_main(struct iwbmpwcontext *wctx)
 	wctx->bmpversion = 0;
 	optv = iw_get_option(wctx->ctx, "bmp:version");
 	if(optv) {
-		wctx->bmpversion = iw_parse_int(optv);
+		if(!strcmp(optv, "auto")) {
+			if(IW_IMGTYPE_HAS_ALPHA(img->imgtype)) {
+				wctx->bmpversion = 5;
+			}
+			else {
+				wctx->bmpversion = 3;
+			}
+		}
+		else {
+			wctx->bmpversion = iw_parse_int(optv);
+		}
 	}
 
 	if(wctx->bmpversion==0) wctx->bmpversion=3;
