@@ -23,13 +23,15 @@
 #endif
 
 struct my_error_mgr {
-	struct jpeg_error_mgr pub;
+	struct jpeg_error_mgr pub; // This field must be first.
+	int have_libjpeg_error;
 	jmp_buf setjmp_buffer;
 };
 
 static void my_error_exit(j_common_ptr cinfo)
 {
 	struct my_error_mgr* myerr = (struct my_error_mgr*)cinfo->err;
+	myerr->have_libjpeg_error = 1;
 	longjmp(myerr->setjmp_buffer, 1);
 }
 
@@ -52,12 +54,36 @@ struct iwjpegrcontext {
 	unsigned int exif_density_unit; // 0 means not set
 };
 
+// TODO?: Combine iwjpegrcontext and jr_rsrc_struct into one struct (or something).
+struct jr_rsrc_struct {
+	struct iw_context *ctx;
+	struct iw_iodescr *iodescr;
+	struct iwjpegrcontext rctx;
+	JSAMPLE *tmprow;
+	struct iw_image img;
+	int cinfo_valid;
+	struct jpeg_decompress_struct cinfo;
+	struct my_error_mgr jerr;
+};
+
 struct iw_exif_state {
 	int endian;
 	const iw_byte *d;
 	size_t d_len;
 };
 
+static unsigned int get_exif_ui16(struct iw_exif_state *e, unsigned int pos)
+{
+	if(e->d_len<2 || pos>e->d_len-2) return 0;
+	return iw_get_ui16_e(&e->d[pos], e->endian);
+}
+
+static unsigned int get_exif_ui32(struct iw_exif_state *e, unsigned int pos)
+{
+	if(e->d_len<4 || pos>e->d_len-4) return 0;
+	return iw_get_ui32_e(&e->d[pos], e->endian);
+}
+
 // Try to read an Exif tag into an integer.
 // Returns zero on failure.
 static int get_exif_tag_int_value(struct iw_exif_state *e, unsigned int tag_pos,
@@ -66,17 +92,17 @@ static int get_exif_tag_int_value(struct iw_exif_state *e, unsigned int tag_pos,
 	unsigned int field_type;
 	unsigned int value_count;
 
-	field_type = iw_get_ui16_e(&e->d[tag_pos+2],e->endian);
-	value_count = iw_get_ui32_e(&e->d[tag_pos+4],e->endian);
+	field_type = get_exif_ui16(e, tag_pos+2);
+	value_count = get_exif_ui32(e, tag_pos+4);
 
 	if(value_count!=1) return 0;
 
 	if(field_type==3) { // SHORT (uint16)
-		*pv = iw_get_ui16_e(&e->d[tag_pos+8],e->endian);
+		*pv = get_exif_ui16(e, tag_pos+8);
 		return 1;
 	}
 	else if(field_type==4) { // LONG (uint32)
-		*pv = iw_get_ui32_e(&e->d[tag_pos+8],e->endian);
+		*pv = get_exif_ui32(e, tag_pos+8);
 		return 1;
 	}
 
@@ -93,8 +119,8 @@ static int get_exif_tag_dbl_value(struct iw_exif_state *e, unsigned int tag_pos,
 	unsigned int value_pos;
 	unsigned int numer, denom;
 
-	field_type = iw_get_ui16_e(&e->d[tag_pos+2],e->endian);
-	value_count = iw_get_ui32_e(&e->d[tag_pos+4],e->endian);
+	field_type = get_exif_ui16(e, tag_pos+2);
+	value_count = get_exif_ui32(e, tag_pos+4);
 
 	if(value_count!=1) return 0;
 
@@ -103,12 +129,12 @@ static int get_exif_tag_dbl_value(struct iw_exif_state *e, unsigned int tag_pos,
 	// A rational is 8 bytes. Since 8>4, it is stored indirectly. First, read
 	// the location where it is stored.
 
-	value_pos = iw_get_ui32_e(&e->d[tag_pos+8],e->endian);
+	value_pos = get_exif_ui32(e, tag_pos+8);
 	if(value_pos > e->d_len-8) return 0;
 
 	// Read the actual value.
-	numer = iw_get_ui32_e(&e->d[value_pos  ],e->endian);
-	denom = iw_get_ui32_e(&e->d[value_pos+4],e->endian);
+	numer = get_exif_ui32(e, value_pos);
+	denom = get_exif_ui32(e, value_pos+4);
 	if(denom==0) return 0;
 
 	*pv = ((double)numer)/denom;
@@ -125,15 +151,15 @@ static void iwjpeg_scan_exif_ifd(struct iwjpegrcontext *rctx,
 	unsigned int v;
 	double v_dbl;
 
-	if(ifd<8 || ifd>e->d_len-18) return;
+	if(ifd<8 || e->d_len<18 || ifd>e->d_len-18) return;
 
-	tag_count = iw_get_ui16_e(&e->d[ifd],e->endian);
+	tag_count = get_exif_ui16(e, ifd);
 	if(tag_count>1000) return; // Sanity check.
 
 	for(i=0;i<tag_count;i++) {
 		tag_pos = ifd+2+i*12;
 		if(tag_pos+12 > e->d_len) return; // Avoid overruns.
-		tag_id = iw_get_ui16_e(&e->d[tag_pos],e->endian);
+		tag_id = get_exif_ui16(e, tag_pos);
 
 		switch(tag_id) {
 		case 274: // 274 = Orientation
@@ -177,7 +203,7 @@ static void iwjpeg_scan_exif(struct iwjpegrcontext *rctx,
 
 	e.endian = d[0]=='I' ? IW_ENDIAN_LITTLE : IW_ENDIAN_BIG;
 
-	ifd = iw_get_ui32_e(&d[4],e.endian);
+	ifd = get_exif_ui32(&e, 4);
 
 	iwjpeg_scan_exif_ifd(rctx,&e,ifd);
 }
@@ -290,50 +316,64 @@ static boolean my_fill_input_buffer_fn(j_decompress_ptr cinfo)
 
 	ret = (*rctx->iodescr->read_fn)(rctx->ctx,rctx->iodescr,
 		rctx->buffer,rctx->buffer_len,&bytesread);
-	if(!ret) return FALSE;
-
+	if((!ret) || (bytesread<1)) {
+		iw_set_error(rctx->ctx, "Unexpected end of file");
+		// Return a fake EOI marker.
+		rctx->buffer[0] = 0xffU;
+		rctx->buffer[1] = 0xd9U;
+		rctx->pub.bytes_in_buffer = 2;
+	}
+	else {
+		rctx->pub.bytes_in_buffer = bytesread;
+	}
 	rctx->pub.next_input_byte = rctx->buffer;
-	rctx->pub.bytes_in_buffer = bytesread;
 
-	if(bytesread<1) return FALSE;
 	return TRUE;
 }
 
-static void my_skip_input_data_fn(j_decompress_ptr cinfo, long num_bytes)
+static void my_skip_input_data_fn(j_decompress_ptr cinfo, long num_bytes_to_skip)
 {
 	struct iwjpegrcontext *rctx = (struct iwjpegrcontext*)cinfo->src;
 	size_t bytes_still_to_skip;
-	size_t nbytes;
 	int ret;
 	size_t bytesread;
 
-	if(num_bytes<=0) return;
-	bytes_still_to_skip = (size_t)num_bytes;
+	// If the skip would leave some valid bytes in the buffer, ...
+	if(num_bytes_to_skip < (long)rctx->pub.bytes_in_buffer) {
+		// ... just move the pointer.
+		rctx->pub.next_input_byte += num_bytes_to_skip;
+		rctx->pub.bytes_in_buffer -= num_bytes_to_skip;
+		return;
+	}
 
+	// Otherwise, read + throw away the necessary number of bytes ...
+	bytes_still_to_skip = (size_t)num_bytes_to_skip - rctx->pub.bytes_in_buffer;
 	while(bytes_still_to_skip>0) {
-		if(rctx->pub.bytes_in_buffer>0) {
-			// There are some bytes in the buffer. Skip up to
-			// 'bytes_still_to_skip' of them.
-			nbytes = rctx->pub.bytes_in_buffer;
-			if(nbytes>bytes_still_to_skip)
-				nbytes = bytes_still_to_skip;
+		size_t bytes_to_read;
 
-			rctx->pub.bytes_in_buffer -= nbytes;
-			rctx->pub.next_input_byte += nbytes;
-			bytes_still_to_skip -= nbytes;
+		bytes_to_read = rctx->buffer_len;
+		if(bytes_to_read > bytes_still_to_skip) {
+			bytes_to_read = bytes_still_to_skip;
 		}
 
-		if(bytes_still_to_skip<1) return;
-
-		// Need to read from the file (or do a seek, but we currently don't
+		// Read from the file (could do a seek instead, but we currently don't
 		// support seeking).
 		ret = (*rctx->iodescr->read_fn)(rctx->ctx,rctx->iodescr,
-			rctx->buffer,rctx->buffer_len,&bytesread);
+			rctx->buffer,bytes_to_read,&bytesread);
 		if(!ret) bytesread=0;
+		if(bytesread==0 || bytesread>bytes_still_to_skip) {
+			// If we couldn't read any data, stop.
+			// (We don't attempt to robustly support input streams for which
+			// not all the data is available yet.)
+			break;
+		}
 
-		rctx->pub.next_input_byte = rctx->buffer;
-		rctx->pub.bytes_in_buffer = bytesread;
+		bytes_still_to_skip -= bytesread;
 	}
+
+	// ... and mark the buffer as empty.
+	rctx->pub.next_input_byte = rctx->buffer;
+	rctx->pub.bytes_in_buffer = 0;
 }
 
 static void my_term_source_fn(j_decompress_ptr cinfo)
@@ -354,85 +394,73 @@ static void convert_cmyk_to_rbg(struct iw_context *ctx, const JSAMPLE *src,
 		r = 1.0 - c*(1.0-k) - k;
 		g = 1.0 - m*(1.0-k) - k;
 		b = 1.0 - y*(1.0-k) - k;
-		if(r<0.0) r=0.0; if(r>1.0) r=1.0;
-		if(g<0.0) g=0.0; if(g>1.0) g=1.0;
-		if(b<0.0) b=0.0; if(b>1.0) b=1.0;
+		if(r<0.0) r=0.0;
+		if(r>1.0) r=1.0;
+		if(g<0.0) g=0.0;
+		if(g>1.0) g=1.0;
+		if(b<0.0) b=0.0;
+		if(b>1.0) b=1.0;
 		dst[3*i+0] = (JSAMPLE)(0.5+255.0*r);
 		dst[3*i+1] = (JSAMPLE)(0.5+255.0*g);
 		dst[3*i+2] = (JSAMPLE)(0.5+255.0*b);
 	}
 }
 
-IW_IMPL(int) iw_read_jpeg_file(struct iw_context *ctx, struct iw_iodescr *iodescr)
+static int iw_read_jpeg_file3(struct jr_rsrc_struct *jr)
 {
+	struct iw_context *ctx = jr->ctx;
+	struct iw_iodescr *iodescr = jr->iodescr;
 	int retval=0;
-	struct jpeg_decompress_struct cinfo;
-	struct my_error_mgr jerr;
-	int cinfo_valid=0;
 	int colorspace;
 	JDIMENSION rownum;
 	JSAMPLE *jsamprow;
 	int numchannels=0;
-	struct iw_image img;
-	struct iwjpegrcontext rctx;
-	JSAMPLE *tmprow = NULL;
 	int cmyk_flag = 0;
+	int ret;
 
-	iw_zeromem(&img,sizeof(struct iw_image));
-	iw_zeromem(&cinfo,sizeof(struct jpeg_decompress_struct));
-	iw_zeromem(&jerr,sizeof(struct my_error_mgr));
-	iw_zeromem(&rctx,sizeof(struct iwjpegrcontext));
-
-	cinfo.err = jpeg_std_error(&jerr.pub);
-	jerr.pub.error_exit = my_error_exit;
-	jerr.pub.output_message = my_output_message;
-
-	if (setjmp(jerr.setjmp_buffer)) {
-		char buffer[JMSG_LENGTH_MAX];
-
-		(*cinfo.err->format_message) ((j_common_ptr)&cinfo, buffer);
-
-		iw_set_errorf(ctx,"libjpeg reports read error: %s",buffer);
-
-		goto done;
-	}
-
-	jpeg_create_decompress(&cinfo);
-	cinfo_valid=1;
+	jpeg_create_decompress(&jr->cinfo);
+	jr->cinfo_valid=1;
 
 	// Set up our custom source manager.
-	rctx.pub.init_source = my_init_source_fn;
-	rctx.pub.fill_input_buffer = my_fill_input_buffer_fn;
-	rctx.pub.skip_input_data = my_skip_input_data_fn;
-	rctx.pub.resync_to_restart = jpeg_resync_to_restart; // libjpeg default
-	rctx.pub.term_source = my_term_source_fn;
-	rctx.ctx = ctx;
-	rctx.iodescr = iodescr;
-	rctx.buffer_len = 32768;
-	rctx.buffer = iw_malloc(ctx, rctx.buffer_len);
-	if(!rctx.buffer) goto done;
-	rctx.exif_density_x = -1.0;
-	rctx.exif_density_y = -1.0;
-	cinfo.src = (struct jpeg_source_mgr*)&rctx;
+	jr->rctx.pub.init_source = my_init_source_fn;
+	jr->rctx.pub.fill_input_buffer = my_fill_input_buffer_fn;
+	jr->rctx.pub.skip_input_data = my_skip_input_data_fn;
+	jr->rctx.pub.resync_to_restart = jpeg_resync_to_restart; // libjpeg default
+	jr->rctx.pub.term_source = my_term_source_fn;
+	jr->rctx.ctx = ctx;
+	jr->rctx.iodescr = iodescr;
+	jr->rctx.buffer_len = 32768;
+	jr->rctx.buffer = iw_malloc(ctx, jr->rctx.buffer_len);
+	if(!jr->rctx.buffer) goto done;
+	jr->rctx.exif_density_x = -1.0;
+	jr->rctx.exif_density_y = -1.0;
+	jr->cinfo.src = (struct jpeg_source_mgr*)&jr->rctx;
 
 	// The lazy way. It would be more efficient to use
 	// jpeg_set_marker_processor(), instead of saving everything to memory.
 	// But libjpeg's marker processing functions have fairly complex
 	// requirements.
-	jpeg_save_markers(&cinfo, 0xe1, 65535);
+	jpeg_save_markers(&jr->cinfo, 0xe1, 65535);
 
-	jpeg_read_header(&cinfo, TRUE);
+	ret = jpeg_read_header(&jr->cinfo, TRUE);
+	if(ret != JPEG_HEADER_OK) {
+		// I don't think this is supposed to be possible, assuming the second
+		// param to jpeg_read_header is TRUE, and our fill_input_buffer
+		// function always returns TRUE.
+		iw_set_error(ctx, "Unexpected libjpeg error");
+		goto done;
+	}
 
-	rctx.is_jfif = cinfo.saw_JFIF_marker;
+	jr->rctx.is_jfif = jr->cinfo.saw_JFIF_marker;
 
-	iwjpeg_read_density(ctx,&img,&cinfo);
+	iwjpeg_read_density(ctx,&jr->img,&jr->cinfo);
 
-	iwjpeg_read_saved_markers(&rctx,&cinfo);
+	iwjpeg_read_saved_markers(&jr->rctx,&jr->cinfo);
 
-	jpeg_start_decompress(&cinfo);
+	jpeg_start_decompress(&jr->cinfo);
 
-	colorspace=cinfo.out_color_space;
-	numchannels=cinfo.output_components;
+	colorspace=jr->cinfo.out_color_space;
+	numchannels=jr->cinfo.output_components;
 
 	// libjpeg will automatically convert YCbCr images to RGB, and YCCK images
 	// to CMYK. That leaves GRAYSCALE, RGB, and CMYK for us to handle.
@@ -440,14 +468,14 @@ IW_IMPL(int) iw_read_jpeg_file(struct iw_context *ctx, struct iw_iodescr *iodesc
 	// cinfo.out_color_space is the colorspace after conversion.
 
 	if(colorspace==JCS_GRAYSCALE && numchannels==1) {
-		img.imgtype = IW_IMGTYPE_GRAY;
-		img.native_grayscale = 1;
+		jr->img.imgtype = IW_IMGTYPE_GRAY;
+		jr->img.native_grayscale = 1;
 	}
 	else if((colorspace==JCS_RGB) && numchannels==3) {
-		img.imgtype = IW_IMGTYPE_RGB;
+		jr->img.imgtype = IW_IMGTYPE_RGB;
 	}
 	else if((colorspace==JCS_CMYK) && numchannels==4) {
-		img.imgtype = IW_IMGTYPE_RGB;
+		jr->img.imgtype = IW_IMGTYPE_RGB;
 		cmyk_flag = 1;
 	}
 	else {
@@ -455,72 +483,111 @@ IW_IMPL(int) iw_read_jpeg_file(struct iw_context *ctx, struct iw_iodescr *iodesc
 		goto done;
 	}
 
-	img.width = cinfo.output_width;
-	img.height = cinfo.output_height;
-	if(!iw_check_image_dimensions(ctx,img.width,img.height)) {
+	jr->img.width = jr->cinfo.output_width;
+	jr->img.height = jr->cinfo.output_height;
+	if(!iw_check_image_dimensions(ctx,jr->img.width,jr->img.height)) {
 		goto done;
 	}
 
-	img.bit_depth = 8;
-	img.bpr = iw_calc_bytesperrow(img.width,img.bit_depth*numchannels);
+	jr->img.bit_depth = 8;
+	jr->img.bpr = iw_calc_bytesperrow(jr->img.width,jr->img.bit_depth*numchannels);
 
-	img.pixels = (iw_byte*)iw_malloc_large(ctx, img.bpr, img.height);
-	if(!img.pixels) {
+	jr->img.pixels = (iw_byte*)iw_malloc_large(ctx, jr->img.bpr, jr->img.height);
+	if(!jr->img.pixels) {
 		goto done;
 	}
 
 	if(cmyk_flag) {
-		tmprow = iw_malloc(ctx,4*img.width);
-		if(!tmprow) goto done;
+		jr->tmprow = iw_malloc(ctx,4*jr->img.width);
+		if(!jr->tmprow) goto done;
 	}
 
-	while(cinfo.output_scanline < cinfo.output_height) {
-		rownum=cinfo.output_scanline;
-		jsamprow = &img.pixels[img.bpr * rownum];
+	while(jr->cinfo.output_scanline < jr->cinfo.output_height) {
+		rownum=jr->cinfo.output_scanline;
+		jsamprow = &jr->img.pixels[jr->img.bpr * rownum];
 		if(cmyk_flag) {
 			// read into tmprow, then convert and copy to img.pixels
-			jpeg_read_scanlines(&cinfo, &tmprow, 1);
-			convert_cmyk_to_rbg(ctx,tmprow,jsamprow,img.width);
+			jpeg_read_scanlines(&jr->cinfo, &jr->tmprow, 1);
+			convert_cmyk_to_rbg(ctx,jr->tmprow,jsamprow,jr->img.width);
 		}
 		else {
 			// read directly into img.pixels
-			jpeg_read_scanlines(&cinfo, &jsamprow, 1);
+			jpeg_read_scanlines(&jr->cinfo, &jsamprow, 1);
 		}
-		if(cinfo.output_scanline<=rownum) {
+		if(jr->cinfo.output_scanline<=rownum) {
 			iw_set_error(ctx,"Error reading JPEG file");
 			goto done;
 		}
 	}
-	jpeg_finish_decompress(&cinfo);
+	jpeg_finish_decompress(&jr->cinfo);
 
-	handle_exif_density(&rctx, &img);
+	handle_exif_density(&jr->rctx, &jr->img);
 
-	iw_set_input_image(ctx, &img);
+	iw_set_input_image(ctx, &jr->img);
 	// The contents of img no longer belong to us.
-	img.pixels = NULL;
+	jr->img.pixels = NULL;
 
-	if(rctx.exif_orientation>=2 && rctx.exif_orientation<=8) {
+	if(jr->rctx.exif_orientation>=2 && jr->rctx.exif_orientation<=8) {
 		static const unsigned int exif_orient_to_transform[9] =
 		   { 0,0, 1,3,2,4,5,7,6 };
 
 		// An Exif marker indicated an unusual image orientation.
 
-		if(rctx.is_jfif) {
-			// The presence of a JFIF marker implies a particular orientation.
-			// If there's also an Exif marker that says something different,
-			// I'm not sure what we're supposed to do.
-			iw_warning(ctx,"JPEG image has an ambiguous orientation");
-		}
-		iw_reorient_image(ctx,exif_orient_to_transform[rctx.exif_orientation]);
+		// Note that if there is also a JFIF segment (jr->rctx.is_jfif), the
+		// orientation is technically ambiguous. But, these days, the usual
+		// practice is to allow Exif to overrule JFIF.
+
+		iw_reorient_image(ctx,exif_orient_to_transform[jr->rctx.exif_orientation]);
 	}
 
 	retval=1;
 
 done:
-	iw_free(ctx, img.pixels);
-	if(cinfo_valid) jpeg_destroy_decompress(&cinfo);
-	if(rctx.buffer) iw_free(ctx,rctx.buffer);
-	if(tmprow) iw_free(ctx,tmprow);
+	// Don't free memory here; do it in iw_read_jpeg_file().
+	return retval;
+}
+
+// This function serves as a target for longjmp(). It shouldn't do much of
+// anything else.
+static int iw_read_jpeg_file2(struct jr_rsrc_struct *jr)
+{
+	if (setjmp(jr->jerr.setjmp_buffer)) {
+		return 0;
+	}
+
+	return iw_read_jpeg_file3(jr);
+}
+
+IW_IMPL(int) iw_read_jpeg_file(struct iw_context *ctx, struct iw_iodescr *iodescr)
+{
+	int retval = 0;
+	struct jr_rsrc_struct *jr = NULL;
+
+	jr = iw_mallocz(ctx, sizeof(struct jr_rsrc_struct));
+	if(!jr) goto done;
+	jr->ctx = ctx;
+	jr->iodescr = iodescr;
+	jr->cinfo.err = jpeg_std_error(&jr->jerr.pub);
+	jr->jerr.pub.error_exit = my_error_exit;
+	jr->jerr.pub.output_message = my_output_message;
+
+	retval = iw_read_jpeg_file2(jr);
+
+done:
+	if(jr) {
+		if(jr->jerr.have_libjpeg_error) {
+			char buffer[JMSG_LENGTH_MAX];
+
+			(*jr->cinfo.err->format_message) ((j_common_ptr)&jr->cinfo, buffer);
+			iw_set_errorf(jr->ctx, "libjpeg reports read error: %s", buffer);
+		}
+
+		if(jr->cinfo_valid) jpeg_destroy_decompress(&jr->cinfo);
+		if(jr->rctx.buffer) iw_free(ctx, jr->rctx.buffer);
+		iw_free(ctx, jr->img.pixels);
+		if(jr->tmprow) iw_free(ctx, jr->tmprow);
+		iw_free(ctx, jr);
+	}
 	return retval;
 }
 
@@ -534,6 +601,16 @@ struct iwjpegwcontext {
 	size_t buffer_len;
 };
 
+struct jw_rsrc_struct {
+	struct iw_context *ctx;
+	struct iw_iodescr *iodescr;
+	struct iwjpegwcontext wctx;
+	JSAMPROW *row_pointers;
+	int compress_created;
+	struct jpeg_compress_struct cinfo;
+	struct my_error_mgr jerr;
+};
+
 static void iwjpg_set_density(struct iw_context *ctx,struct jpeg_compress_struct *cinfo,
 	const struct iw_image *img)
 {
@@ -597,30 +674,23 @@ static void my_term_destination_fn(j_compress_ptr cinfo)
 	}
 }
 
-IW_IMPL(int) iw_write_jpeg_file(struct iw_context *ctx,  struct iw_iodescr *iodescr)
+static int iw_write_jpeg_file3(struct jw_rsrc_struct *jw)
 {
+	struct iw_context *ctx = jw->ctx;
+	struct iw_iodescr *iodescr = jw->iodescr;
 	int retval=0;
-	struct jpeg_compress_struct cinfo;
-	struct my_error_mgr jerr;
 	J_COLOR_SPACE in_colortype; // Color type of the data we give to libjpeg
 	int jpeg_cmpts;
-	int compress_created = 0;
 	int compress_started = 0;
-	JSAMPROW *row_pointers = NULL;
 	int is_grayscale;
 	int j;
 	struct iw_image img;
 	int jpeg_quality;
 	int samp_factor_h, samp_factor_v;
 	int disable_subsampling = 0;
-	struct iwjpegwcontext wctx;
 	const char *optv;
 	int ret;
 
-	iw_zeromem(&cinfo,sizeof(struct jpeg_compress_struct));
-	iw_zeromem(&jerr,sizeof(struct my_error_mgr));
-	iw_zeromem(&wctx,sizeof(struct iwjpegwcontext));
-
 	iw_get_output_image(ctx,&img);
 
 	if(IW_IMGTYPE_HAS_ALPHA(img.imgtype)) {
@@ -644,41 +714,37 @@ IW_IMPL(int) iw_write_jpeg_file(struct iw_context *ctx,  struct iw_iodescr *iode
 		jpeg_cmpts=3;
 	}
 
-	cinfo.err = jpeg_std_error(&jerr.pub);
-	jerr.pub.error_exit = my_error_exit;
-
-	if (setjmp(jerr.setjmp_buffer)) {
-		char buffer[JMSG_LENGTH_MAX];
-
-		(*cinfo.err->format_message) ((j_common_ptr)&cinfo, buffer);
-
-		iw_set_errorf(ctx,"libjpeg reports write error: %s",buffer);
-
-		goto done;
-	}
-
-	jpeg_create_compress(&cinfo);
-	compress_created=1;
+	jpeg_create_compress(&jw->cinfo);
+	jw->compress_created=1;
 
 	// Set up our custom destination manager.
-	wctx.pub.init_destination = my_init_destination_fn;
-	wctx.pub.empty_output_buffer = my_empty_output_buffer_fn;
-	wctx.pub.term_destination = my_term_destination_fn;
-	wctx.ctx = ctx;
-	wctx.iodescr = iodescr;
-	wctx.buffer_len = 32768;
-	wctx.buffer = iw_malloc(ctx,wctx.buffer_len);
-	if(!wctx.buffer) goto done;
+	jw->wctx.pub.init_destination = my_init_destination_fn;
+	jw->wctx.pub.empty_output_buffer = my_empty_output_buffer_fn;
+	jw->wctx.pub.term_destination = my_term_destination_fn;
+	jw->wctx.ctx = ctx;
+	jw->wctx.iodescr = iodescr;
+	jw->wctx.buffer_len = 32768;
+	jw->wctx.buffer = iw_malloc(ctx,jw->wctx.buffer_len);
+	if(!jw->wctx.buffer) goto done;
 	// Our wctx is organized so it can double as a
 	// 'struct jpeg_destination_mgr'.
-	cinfo.dest = (struct jpeg_destination_mgr*)&wctx;
+	jw->cinfo.dest = (struct jpeg_destination_mgr*)&jw->wctx;
 
-	cinfo.image_width = img.width;
-	cinfo.image_height = img.height;
-	cinfo.input_components = jpeg_cmpts;
-	cinfo.in_color_space = in_colortype;
+	jw->cinfo.image_width = img.width;
+	jw->cinfo.image_height = img.height;
+	jw->cinfo.input_components = jpeg_cmpts;
+	jw->cinfo.in_color_space = in_colortype;
 
-	jpeg_set_defaults(&cinfo);
+	jpeg_set_defaults(&jw->cinfo);
+
+	optv = iw_get_option(ctx, "jpeg:rstm");
+	if(optv) {
+		jw->cinfo.restart_interval = (unsigned int)iw_parse_int(optv);
+	}
+	optv = iw_get_option(ctx, "jpeg:rstr");
+	if(optv) {
+		jw->cinfo.restart_in_rows = (int)iw_parse_int(optv);
+	}
 
 	optv = iw_get_option(ctx, "jpeg:block");
 	if(optv) {
@@ -687,7 +753,7 @@ IW_IMPL(int) iw_write_jpeg_file(struct iw_context *ctx,  struct iw_iodescr *iode
 		// Note: This might not work if DCT_SCALING_SUPPORTED was not defined when
 		// libjpeg was compiled, but that symbol is not normally exposed to
 		// applications.
-		cinfo.block_size = iw_parse_int(optv);
+		jw->cinfo.block_size = iw_parse_int(optv);
 #else
 		iw_warning(ctx, "Setting block size is not supported by this version of libjpeg");
 #endif
@@ -695,42 +761,47 @@ IW_IMPL(int) iw_write_jpeg_file(struct iw_context *ctx,  struct iw_iodescr *iode
 
 	optv = iw_get_option(ctx, "jpeg:arith");
 	if(optv)
-		cinfo.arith_code = iw_parse_int(optv) ? TRUE : FALSE;
+		jw->cinfo.arith_code = iw_parse_int(optv) ? TRUE : FALSE;
 	else
-		cinfo.arith_code = FALSE;
+		jw->cinfo.arith_code = FALSE;
 
 	optv = iw_get_option(ctx, "jpeg:colortype");
 	if(optv) {
 		if(!strcmp(optv, "rgb")) {
 			if(in_colortype==JCS_RGB) {
-				jpeg_set_colorspace(&cinfo,JCS_RGB);
+				jpeg_set_colorspace(&jw->cinfo,JCS_RGB);
 				disable_subsampling = 1;
 			}
 		}
 		else if(!strcmp(optv, "rgb1")) {
 			if(in_colortype==JCS_RGB) {
 #if JPEG_LIB_VERSION_MAJOR >= 9
-				cinfo.color_transform = JCT_SUBTRACT_GREEN;
+				jw->cinfo.color_transform = JCT_SUBTRACT_GREEN;
 #else
 				iw_warning(ctx, "Color type rgb1 is not supported by this version of libjpeg");
 #endif
-				jpeg_set_colorspace(&cinfo,JCS_RGB);
+				jpeg_set_colorspace(&jw->cinfo,JCS_RGB);
 				disable_subsampling = 1;
 			}
 		}
 	}
 
+	optv = iw_get_option(ctx, "jpeg:optcoding");
+	if(optv && iw_parse_int(optv)) {
+		jw->cinfo.optimize_coding = TRUE;
+	}
+
 	optv = iw_get_option(ctx, "jpeg:bgycc");
 	if(optv && iw_parse_int(optv)) {
 #if (JPEG_LIB_VERSION_MAJOR>9 || \
 	(JPEG_LIB_VERSION_MAJOR==9 && JPEG_LIB_VERSION_MINOR>=1))
-		jpeg_set_colorspace(&cinfo, JCS_BG_YCC);
+		jpeg_set_colorspace(&jw->cinfo, JCS_BG_YCC);
 #else
 		iw_warning(ctx, "Big gamut YCC is not supported by this version of libjpeg");
 #endif
 	}
 
-	iwjpg_set_density(ctx,&cinfo,&img);
+	iwjpg_set_density(ctx,&jw->cinfo,&img);
 
 	optv = iw_get_option(ctx, "jpeg:quality");
 	if(optv)
@@ -739,7 +810,7 @@ IW_IMPL(int) iw_write_jpeg_file(struct iw_context *ctx,  struct iw_iodescr *iode
 		jpeg_quality = 0;
 
 	if(jpeg_quality>0) {
-		jpeg_set_quality(&cinfo,jpeg_quality,0);
+		jpeg_set_quality(&jw->cinfo,jpeg_quality,0);
 	}
 
 	if(jpeg_cmpts>1 && !disable_subsampling) {
@@ -773,43 +844,81 @@ IW_IMPL(int) iw_write_jpeg_file(struct iw_context *ctx,  struct iw_iodescr *iode
 
 		if(samp_factor_h>0) {
 			if(samp_factor_h>4) samp_factor_h=4;
-			cinfo.comp_info[0].h_samp_factor = samp_factor_h;
+			jw->cinfo.comp_info[0].h_samp_factor = samp_factor_h;
 		}
 		if(samp_factor_v>0) {
 			if(samp_factor_v>4) samp_factor_v=4;
-			cinfo.comp_info[0].v_samp_factor = samp_factor_v;
+			jw->cinfo.comp_info[0].v_samp_factor = samp_factor_v;
 		}
 	}
 
 	if(iw_get_value(ctx,IW_VAL_OUTPUT_INTERLACED)) {
-		jpeg_simple_progression(&cinfo);
+		jpeg_simple_progression(&jw->cinfo);
 	}
 
-	row_pointers = (JSAMPROW*)iw_malloc(ctx, img.height * sizeof(JSAMPROW));
-	if(!row_pointers) goto done;
+	jw->row_pointers = (JSAMPROW*)iw_malloc(ctx, img.height * sizeof(JSAMPROW));
+	if(!jw->row_pointers) goto done;
 
 	for(j=0;j<img.height;j++) {
-		row_pointers[j] = &img.pixels[j*img.bpr];
+		jw->row_pointers[j] = &img.pixels[j*img.bpr];
 	}
 
-	jpeg_start_compress(&cinfo, TRUE);
+	jpeg_start_compress(&jw->cinfo, TRUE);
 	compress_started=1;
 
-	jpeg_write_scanlines(&cinfo, row_pointers, img.height);
+	jpeg_write_scanlines(&jw->cinfo, jw->row_pointers, img.height);
 
 	retval=1;
 
 done:
 	if(compress_started)
-		jpeg_finish_compress(&cinfo);
+		jpeg_finish_compress(&jw->cinfo);
 
-	if(compress_created)
-		jpeg_destroy_compress(&cinfo);
+	// Don't free memory here; do it in iw_write_jpeg_file().
+	return retval;
+}
 
-	if(row_pointers) iw_free(ctx,row_pointers);
+// This function serves as a target for longjmp(). It shouldn't do much of
+// anything else.
+static int iw_write_jpeg_file2(struct jw_rsrc_struct *jw)
+{
+	if (setjmp(jw->jerr.setjmp_buffer)) {
+		return 0;
+	}
 
-	if(wctx.buffer) iw_free(ctx,wctx.buffer);
+	return iw_write_jpeg_file3(jw);
+}
 
+IW_IMPL(int) iw_write_jpeg_file(struct iw_context *ctx,  struct iw_iodescr *iodescr)
+{
+	struct jw_rsrc_struct *jw = NULL;
+	int retval = 0;
+
+	jw = iw_mallocz(ctx, sizeof(struct jw_rsrc_struct));
+	if(!jw) goto done;
+	jw->ctx = ctx;
+	jw->iodescr = iodescr;
+
+	jw->cinfo.err = jpeg_std_error(&jw->jerr.pub);
+	jw->jerr.pub.error_exit = my_error_exit;
+	jw->jerr.pub.output_message = my_output_message;
+
+	retval = iw_write_jpeg_file2(jw);
+
+done:
+	if(jw) {
+		if(jw->jerr.have_libjpeg_error) {
+			char buffer[JMSG_LENGTH_MAX];
+
+			(*jw->cinfo.err->format_message) ((j_common_ptr)&jw->cinfo, buffer);
+			iw_set_errorf(jw->ctx, "libjpeg reports write error: %s", buffer);
+		}
+
+		if(jw->compress_created) jpeg_destroy_compress(&jw->cinfo);
+		if(jw->row_pointers) iw_free(ctx, jw->row_pointers);
+		if(jw->wctx.buffer) iw_free(ctx, jw->wctx.buffer);
+		iw_free(ctx, jw);
+	}
 	return retval;
 }
 
